#define G_LOG_DOMAIN "ap-ability"

#include "ap-ability.h"

#include <math.h>
#include <glib.h>
#include <glib-object.h>

typedef struct _APAbilityPrivate
{
    gchar *name;
    gint   score;
    gint   modifier;

    gint   base;
    gint   racial_bonus;
    gint   ability_improvements;
    gint   miscellaneous_bonus;
    gint   set_score;
    gint   other_modifier;

    gint   override_score;
} APAbilityPrivate;

struct _APAbility
{
    GObject parent_instance;
};

/**
 * SECTION:ap-ability
 * @short_description: Manages a single ability of a character
 * @title: APAbility
 * @include: adventure-pack/ap-ability.h
 *
 * #APAbility is an object for creating and managing a character's ability.
 * Allows the setting of various ability attributes and automatically calculate
 * values like modifiers.
 *
 * # Usage {#usage}
 * ## Basic
 * |[<!-- language="C" -->
 * APAbility *ability;
 *
 * ability = g_object_new (AP_TYPE_ABILITY,
 *                         "name", "Dexterity",
 *                         "base", 10,
 *                         NULL);
 *
 * g_print ("%s: %i (%i)",
 *          ap_ability_get_name (ability),
 *          ap_ability_get_score (ability),
 *          ap_ability_get_modifier (ability));
 *
 * // Returns: Dexterity: 10 (0)
 * ]|
 *
 * ## Add properties
 * |[<!-- language="C" -->
 * APAbility *ability;
 *
 * ability = ap_ability_new_with_name ("Dexterity");
 * ap_ability_add_score (ability, 10);
 * ]|
 */

G_DEFINE_TYPE_WITH_PRIVATE (APAbility, ap_ability, G_TYPE_OBJECT)

enum
{
    PROP_0,
    PROP_NAME,
    PROP_SCORE,
    PROP_BASE,
    PROP_MODIFER,
    PROP_RACIAL_BONUS,
    PROP_ABILITY_IMPROVEMENTS,
    PROP_MISCELLANEOUS_BONUS,
    PROP_SET_SCORE,
    PROP_OTHER_MODIFIER,
    PROP_OVERRIDE_SCORE,
    NUM_PROPERTIES
};

static GParamSpec *properties[NUM_PROPERTIES] = { NULL, };

static void
calculate_modifier (APAbility *self)
{
    APAbilityPrivate *priv = ap_ability_get_instance_private (self);
    gint score;
    gint modifier;

    /*
     * When a score number will result in a negative number (ex. 9 - 10 = -1)
     * then we subtract by 1 more to return a proper floor ().
     */
    if (priv->score < 10)
        score = priv->score - 1;
    else
        score = priv->score;

    /*
     * Modifier = ((Ability score) - 10) / 2
     */
    modifier = floor ((score - 10) / 2);

    if (modifier == priv->modifier)
        return;

    priv->modifier = modifier;

    g_object_notify_by_pspec (G_OBJECT (self), properties[PROP_MODIFER]);
}

static void
calculate_score (APAbility *self)
{
    APAbilityPrivate *priv = ap_ability_get_instance_private (self);
    gint score;

    if (priv->override_score)
    {
        score = priv->override_score;
    }
    else
    {
        score = priv->base +
                priv->modifier +
                priv->racial_bonus +
                priv->ability_improvements +
                priv->miscellaneous_bonus +
                priv->set_score +
                priv->other_modifier;
    }


    if (score == priv->score)
        return;

    priv->score = score;

    g_object_notify_by_pspec (G_OBJECT (self), properties[PROP_SCORE]);

    /* Recalculate modifier */
    calculate_modifier (self);
}

static void
ap_ability_set_property (GObject      *object,
                         guint         property_id,
                         const GValue *value,
                         GParamSpec   *pspec)
{
    APAbility *self = AP_ABILITY (object);
    APAbilityPrivate *priv = ap_ability_get_instance_private (self);

    switch (property_id)
    {
        case PROP_NAME:
            priv->name = g_value_dup_string (value);
            break;

        case PROP_BASE:
            priv->base = g_value_get_int (value);
            calculate_score (self);
            break;

        case PROP_RACIAL_BONUS:
            priv->racial_bonus = g_value_get_int (value);
            calculate_score (self);
            break;

        case PROP_ABILITY_IMPROVEMENTS:
            priv->ability_improvements = g_value_get_int (value);
            calculate_score (self);
            break;

        case PROP_MISCELLANEOUS_BONUS:
            priv->miscellaneous_bonus = g_value_get_int (value);
            calculate_score (self);
            break;

        case PROP_SET_SCORE:
            priv->set_score = g_value_get_int (value);
            calculate_score (self);
            break;

        case PROP_OTHER_MODIFIER:
            priv->other_modifier = g_value_get_int (value);
            calculate_score (self);
            break;

        case PROP_OVERRIDE_SCORE:
            priv->override_score = g_value_get_int (value);
            calculate_score (self);
            break;

        default:
            G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
            break;
    }
}

static void
ap_ability_get_property (GObject      *object,
                         guint         property_id,
                         GValue       *value,
                         GParamSpec   *pspec)
{
    APAbility *self = AP_ABILITY (object);
    APAbilityPrivate *priv = ap_ability_get_instance_private (self);

    switch (property_id)
    {
        case PROP_NAME:
            g_value_set_string (value, priv->name);
            break;

        case PROP_SCORE:
            g_value_set_int (value, priv->score);
            break;

        case PROP_MODIFER:
            g_value_set_int (value, priv->modifier);
            break;

        case PROP_RACIAL_BONUS:
            g_value_set_int (value, priv->racial_bonus);
            break;

        case PROP_ABILITY_IMPROVEMENTS:
            g_value_set_int (value, priv->ability_improvements);
            break;

        case PROP_MISCELLANEOUS_BONUS:
            g_value_set_int (value, priv->miscellaneous_bonus);
            break;

        case PROP_SET_SCORE:
            g_value_set_int (value, priv->set_score);
            break;

        case PROP_OTHER_MODIFIER:
            g_value_set_int (value, priv->other_modifier);
            break;

        case PROP_OVERRIDE_SCORE:
            g_value_set_int (value, priv->override_score);
            break;

        default:
            G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
            break;
    }
}

static void
ap_ability_finalize (GObject *object)
{
    APAbility *self = AP_ABILITY (object);
    APAbilityPrivate *priv = ap_ability_get_instance_private (self);

    g_free (priv->name);

    G_OBJECT_CLASS (ap_ability_parent_class)->finalize (object);
}

static void
ap_ability_init (APAbility *self)
{
    APAbilityPrivate *priv = ap_ability_get_instance_private (self);

    priv->name = NULL;
    priv->score = 0;
    priv->modifier = 0;
    priv->base = 0;
    priv->racial_bonus = 0;
    priv->ability_improvements = 0;
    priv->miscellaneous_bonus = 0;
    priv->set_score = 0;
    priv->other_modifier = 0;
    priv->override_score = 0;
}

static void
ap_ability_class_init (APAbilityClass *klass)
{
    GObjectClass *object_class = G_OBJECT_CLASS (klass);

    object_class->set_property = ap_ability_set_property;
    object_class->get_property = ap_ability_get_property;
    object_class->finalize = ap_ability_finalize;

	/**
	 * APAbility:name:
	 *
	 * Name of the ability.
     *
     * # Example
     * * `Dexterity`
	 */
    properties[PROP_NAME] =
        g_param_spec_string ("name",
                             "Name",
                             "Name of the ability.",
                             NULL,
                             G_PARAM_READWRITE);

	/**
	 * APAbility:score:
	 *
	 * Ability score level.
	 */
    properties[PROP_SCORE] =
        g_param_spec_int ("score",
                          "Score",
                          "Score of the ability.",
                          G_MININT,
                          G_MAXINT,
                          0,
                          G_PARAM_READABLE);

	/**
	 * APAbility:modifier:
	 *
	 * (Read-only): Returns the ability modifier
     *
     * modifier = ((#APAbility:score) - 10) / 2
	 */
    properties[PROP_MODIFER] =
        g_param_spec_int ("modifier",
                          "Modifier",
                          "Modifier based of the ability score.",
                          G_MININT,
                          G_MAXINT,
                          0,
                          G_PARAM_READABLE);

	/**
	 * APAbility:base:
	 *
	 * Ability base score level.
	 */
    properties[PROP_BASE] =
        g_param_spec_int ("base",
                          "Base",
                          "Base score of the ability.",
                          G_MININT,
                          G_MAXINT,
                          0,
                          G_PARAM_READWRITE);

	/**
	 * APAbility:racial-bonus:
	 *
	 * Ability score bonus.
	 */
    properties[PROP_RACIAL_BONUS] =
        g_param_spec_int ("racial-bonus",
                          "Racial bonus",
                          "Racial bonus score of the ability.",
                          G_MININT,
                          G_MAXINT,
                          0,
                          G_PARAM_READWRITE);

	/**
	 * APAbility:ability-improvements:
	 *
	 * Ability improvements.
	 */
    properties[PROP_ABILITY_IMPROVEMENTS] =
        g_param_spec_int ("ability-improvements",
                          "Ability improvements",
                          "Ability improvements of the ability.",
                          G_MININT,
                          G_MAXINT,
                          0,
                          G_PARAM_READWRITE);

	/**
	 * APAbility:miscellaneous-bonus:
	 *
	 * Any miscellaneous ability bonus.
	 */
    properties[PROP_MISCELLANEOUS_BONUS] =
        g_param_spec_int ("miscellaneous-bonus",
                          "Miscellaneous bonus",
                          "Miscellaneous bonus of the ability.",
                          G_MININT,
                          G_MAXINT,
                          0,
                          G_PARAM_READWRITE);

	/**
	 * APAbility:set-score:
	 *
	 * Ability set score.
	 */
    properties[PROP_SET_SCORE] =
        g_param_spec_int ("set-score",
                          "Set score",
                          "Set score of the ability.",
                          G_MININT,
                          G_MAXINT,
                          0,
                          G_PARAM_READWRITE);

	/**
	 * APAbility:other-modifier:
	 *
	 * Other ability modifier, if one.
	 */
    properties[PROP_OTHER_MODIFIER] =
        g_param_spec_int ("other-modifier",
                          "Other modifier",
                          "Other ability modifier.",
                          G_MININT,
                          G_MAXINT,
                          0,
                          G_PARAM_READWRITE);

	/**
	 * APAbility:override-score:
	 *
	 * Regardless of the other properties the ability score will be set to this
     * value.
	 */
    properties[PROP_OVERRIDE_SCORE] =
        g_param_spec_int ("override",
                          "Override",
                          "Override ability scorey.",
                          G_MININT,
                          G_MAXINT,
                          0,
                          G_PARAM_READWRITE);


    g_object_class_install_properties (object_class,
                                       NUM_PROPERTIES,
                                       properties);
}

/*
 * *****************************************************************************
 * API
 * *****************************************************************************
 */
/**
 * ap_ability_new:
 *
 * Creates a new #APAbility.
 *
 * Returns: (transfer full): an #APAbility instance
 */
APAbility *
ap_ability_new (void)
{
    return g_object_new (AP_TYPE_ABILITY, NULL);
}

/**
 * ap_ability_new_with_name:
 * @name: the ability name for new #APAbility
 *
 * Creates a new #APAbility with @name.
 *
 * Returns: (transfer full): an #APAbility instance
 */
APAbility *
ap_ability_new_with_name (gchar *name)
{
    return g_object_new (AP_TYPE_ABILITY,
                         "name", name,
                         NULL);
}

/**
 * ap_ability_set_name:
 * @self: an #APAbility
 * @name: the ability name for @self
 *
 * Sets the ability name for @self.
 *
 * ## Examples
 * * `Strength`
 * * `Dexterity`
 */
void
ap_ability_set_name (APAbility *self,
                     gchar     *name)
{
    GValue new_name = G_VALUE_INIT;

    g_assert (AP_IS_ABILITY (self));
    g_assert (name != NULL);
    g_assert (g_strcmp0 (name, "") != 0);

    g_value_init (&new_name, G_TYPE_STRING);
    g_value_set_string (&new_name, name);

    g_object_set_property (G_OBJECT (self), "name", &new_name);

    g_value_unset (&new_name);
}

/**
 * ap_ability_get_name:
 * @self: an #APAbility
 *
 * Gets the name of @self.
 *
 * Returns: (transfer full) (nullable): the ability name of @self, free with
 *          g_free()
 */
gchar *
ap_ability_get_name (APAbility *self)
{
    GValue val = G_VALUE_INIT;
    gchar *retval;

    g_assert (AP_IS_ABILITY (self));

    g_value_init (&val, G_TYPE_STRING);

    g_object_get_property (G_OBJECT (self), "name", &val);

    retval = g_value_dup_string (&val);

    g_value_unset (&val);

    return retval;
}

/**
 * ap_ability_get_score:
 * @self: an #APAbility
 *
 * Gets the ability score of @self.
 *
 * Returns: (nullable): the ability score of @self
 */
gint
ap_ability_get_score (APAbility *self)
{
    GValue val = G_VALUE_INIT;
    gint retval;

    g_assert (AP_IS_ABILITY (self));

    g_value_init (&val, G_TYPE_INT);

    g_object_get_property (G_OBJECT (self), "score", &val);

    retval = g_value_get_int (&val);

    g_value_unset (&val);

    return retval;
}

/**
 * ap_ability_get_modifier:
 * @self: an #APAbility
 *
 * Gets the ability score modifier of @self.
 *
 * Returns: (nullable): the ability score modifier of @self
 */
gint
ap_ability_get_modifier (APAbility *self)
{
    GValue val = G_VALUE_INIT;
    gint retval;

    g_assert (AP_IS_ABILITY (self));

    g_value_init (&val, G_TYPE_INT);

    g_object_get_property (G_OBJECT (self), "modifier", &val);

    retval = g_value_get_int (&val);

    g_value_unset (&val);

    return retval;
}

/**
 * ap_ability_set_base:
 * @self: an #APAbility
 * @base: the base ability score for @self
 *
 * Sets the base ability score for @self.
 */
void
ap_ability_set_base (APAbility *self,
                     gint       base)
{
    GValue new_base = G_VALUE_INIT;

    g_assert (AP_IS_ABILITY (self));

    g_value_init (&new_base, G_TYPE_INT);
    g_value_set_int (&new_base, base);

    g_object_set_property (G_OBJECT (self), "base", &new_base);

    g_value_unset (&new_base);
}

/**
 * ap_ability_get_base:
 * @self: an #APAbility
 *
 * Gets the base ability score of @self.
 *
 * Returns: (nullable): the base ability score of @self
 */
gint
ap_ability_get_base (APAbility *self)
{
    GValue val = G_VALUE_INIT;
    gint retval;

    g_assert (AP_IS_ABILITY (self));

    g_value_init (&val, G_TYPE_INT);

    g_object_get_property (G_OBJECT (self), "base", &val);

    retval = g_value_get_int (&val);

    g_value_unset (&val);

    return retval;
}

/**
 * ap_ability_set_bonus:
 * @self: an #APAbility
 * @racial_bonus: the racial bonus for the @self
 *
 * Sets the racial bonus for @self.
 */
void
ap_ability_set_racial_bonus (APAbility *self,
                             gint       racial_bonus)
{
    GValue new_racial_bonus = G_VALUE_INIT;

    g_assert (AP_IS_ABILITY (self));

    g_value_init (&new_racial_bonus, G_TYPE_INT);
    g_value_set_int (&new_racial_bonus, racial_bonus);

    g_object_set_property (G_OBJECT (self), "racial-bonus", &new_racial_bonus);

    g_value_unset (&new_racial_bonus);
}

/**
 * ap_ability_get_racial_bonus:
 * @self: an #APAbility
 *
 * Gets the racial bonus of @self.
 *
 * Returns: (nullable): the racial bonus of @self
 */
gint
ap_ability_get_racial_bonus (APAbility *self)
{
    GValue val = G_VALUE_INIT;
    gint retval;

    g_assert (AP_IS_ABILITY (self));

    g_value_init (&val, G_TYPE_INT);

    g_object_get_property (G_OBJECT (self), "racial-bonus", &val);

    retval = g_value_get_int (&val);

    g_value_unset (&val);

    return retval;
}

/**
 * ap_ability_ability_improvements:
 * @self: an #APAbility
 * @ability_improvements: the ability improvements for the @self
 *
 * Sets the ability improvements for @self.
 */
void
ap_ability_set_ability_improvements (APAbility *self,
                                     gint       ability_improvements)
{
    GValue new_ability_improvements = G_VALUE_INIT;

    g_assert (AP_IS_ABILITY (self));

    g_value_init (&new_ability_improvements, G_TYPE_INT);
    g_value_set_int (&new_ability_improvements, ability_improvements);

    g_object_set_property (G_OBJECT (self), "ability-improvements", &new_ability_improvements);

    g_value_unset (&new_ability_improvements);
}

/**
 * ap_ability_get_ability_improvements:
 * @self: an #APAbility
 *
 * Gets the ability improvements of @self.
 *
 * Returns: (nullable): the ability improvements of @self
 */
gint
ap_ability_get_ability_improvements (APAbility *self)
{
    GValue val = G_VALUE_INIT;
    gint retval;

    g_assert (AP_IS_ABILITY (self));

    g_value_init (&val, G_TYPE_INT);

    g_object_get_property (G_OBJECT (self), "ability-improvements", &val);

    retval = g_value_get_int (&val);

    g_value_unset (&val);

    return retval;
}

/**
 * ap_ability_miscellaneous_bonus:
 * @self: an #APAbility
 * @miscellaneous_bonus: the miscellaneous bonus for the @self
 *
 * Sets the miscellaneous bonus for @self.
 */
void
ap_ability_set_miscellaneous_bonus (APAbility *self,
                                    gint       miscellaneous_bonus)
{
    GValue new_miscellaneous_bonus = G_VALUE_INIT;

    g_assert (AP_IS_ABILITY (self));

    g_value_init (&new_miscellaneous_bonus, G_TYPE_INT);
    g_value_set_int (&new_miscellaneous_bonus, miscellaneous_bonus);

    g_object_set_property (G_OBJECT (self), "miscellaneous-bonus", &new_miscellaneous_bonus);

    g_value_unset (&new_miscellaneous_bonus);
}

/**
 * ap_ability_get_miscellaneous_bonus:
 * @self: an #APAbility
 *
 * Gets the miscellaneous bonus of @self.
 *
 * Returns: (nullable): the miscellaneous bonus of @self
 */
gint
ap_ability_get_miscellaneous_bonus (APAbility *self)
{
    GValue val = G_VALUE_INIT;
    gint retval;

    g_assert (AP_IS_ABILITY (self));

    g_value_init (&val, G_TYPE_INT);

    g_object_get_property (G_OBJECT (self), "miscellaneous-bonus", &val);

    retval = g_value_get_int (&val);

    g_value_unset (&val);

    return retval;
}

/**
 * ap_ability_set_other_modifier:
 * @self: an #APAbility
 * @other_modifier: the other modifier for the @self
 *
 * Sets the other modifier for @self.
 */
void
ap_ability_set_other_modifier (APAbility *self,
                               gint       other_modifier)
{
    GValue new_other_modifier = G_VALUE_INIT;

    g_assert (AP_IS_ABILITY (self));

    g_value_init (&new_other_modifier, G_TYPE_INT);
    g_value_set_int (&new_other_modifier, other_modifier);

    g_object_set_property (G_OBJECT (self), "other-modifier", &new_other_modifier);

    g_value_unset (&new_other_modifier);
}

/**
 * ap_ability_get_other_modifier:
 * @self: an #APAbility
 *
 * Gets the other modifier of @self.
 *
 * Returns: (nullable): the other modifier of @self
 */
gint
ap_ability_get_other_modifier (APAbility *self)
{
    GValue val = G_VALUE_INIT;
    gint retval;

    g_assert (AP_IS_ABILITY (self));

    g_value_init (&val, G_TYPE_INT);

    g_object_get_property (G_OBJECT (self), "other-modifier", &val);

    retval = g_value_get_int (&val);

    g_value_unset (&val);

    return retval;
}

/**
 * ap_ability_set_override_score:
 * @self: an #APAbility
 * @override_score: the override score for the @self
 *
 * Sets the override score for @self.
 */
void
ap_ability_set_override_score (APAbility *self,
                               gint       override_score)
{
    GValue new_override_score = G_VALUE_INIT;

    g_assert (AP_IS_ABILITY (self));

    g_value_init (&new_override_score, G_TYPE_INT);
    g_value_set_int (&new_override_score, override_score);

    g_object_set_property (G_OBJECT (self), "override-score", &new_override_score);

    g_value_unset (&new_override_score);
}

/**
 * ap_ability_get_override_score:
 * @self: an #APAbility
 *
 * Gets the override score of @self.
 *
 * Returns: (nullable): the override score of @self
 */
gint
ap_ability_get_override_score (APAbility *self)
{
    GValue val = G_VALUE_INIT;
    gint retval;

    g_assert (AP_IS_ABILITY (self));

    g_value_init (&val, G_TYPE_INT);

    g_object_get_property (G_OBJECT (self), "override-score", &val);

    retval = g_value_get_int (&val);

    g_value_unset (&val);

    return retval;
}
