#include "test-character.h"

#include <adventure-pack/ap-ability.h>
#include <adventure-pack/ap-skill.h>
#include <adventure-pack/ap-character.h>

#include <glib.h>

void
character_fixture_set_up (CharacterFixture *fixture,
                          gconstpointer     user_data)
{
    fixture->character = ap_character_new ();

}

void
character_fixture_tear_down (CharacterFixture *fixture,
                             gconstpointer     user_data)
{
    g_clear_object (&fixture->character);
}

void
test_character_ability_add (CharacterFixture *fixture,
                            gconstpointer     user_data)
{
    APAbility *ability;
    APAbility *ability_from_character;
    gboolean retval;

    ability = ap_ability_new_with_name ("Dexterity");

    retval = ap_character_add_ability (fixture->character, ability);

    g_assert (retval == TRUE);

    ability_from_character = ap_character_get_ability (fixture->character, "Dexterity");

    g_assert (ability_from_character == ability);
}

void
test_character_ability_add_with_no_name (CharacterFixture *fixture,
                                         gconstpointer     user_data)
{
    APAbility *ability;

    ability = ap_ability_new ();

    g_test_expect_message ("ap-character",
                           G_LOG_LEVEL_WARNING,
                           "Ability not registered, ability must have a name to be registered.");

    ap_character_add_ability (fixture->character, ability);

    g_test_assert_expected_messages ();
}

void
test_character_ability_add_with_duplicate_name (CharacterFixture *fixture,
                                                gconstpointer     user_data)
{
    APAbility *ability;
    gboolean retval;

    ability = ap_ability_new_with_name ("Dexterity");
    retval = ap_character_add_ability (fixture->character, ability);
    g_assert (retval == TRUE);

    ability = ap_ability_new_with_name ("Dexterity");

    g_test_expect_message ("ap-character",
                           G_LOG_LEVEL_WARNING,
                           "Ability not registered, ability with name \"*\" already exists.");

    ap_character_add_ability (fixture->character, ability);

    g_test_assert_expected_messages ();
}

void
test_character_ability_get (CharacterFixture *fixture,
                            gconstpointer     user_data)
{
    APAbility *ability;
    APAbility *ability_from_character;

    ability = ap_ability_new ();

    ap_ability_set_name (ability, "Dexterity");

    ap_character_add_ability (fixture->character, ability);

    ability_from_character = ap_character_get_ability (fixture->character, "Dexterity");

    g_assert (AP_IS_ABILITY (ability_from_character));
    g_assert (ability == ability_from_character);
}

void
test_character_skill_add (CharacterFixture *fixture,
                          gconstpointer     user_data)
{
    APSkill *skill;
    gboolean retval;

    skill = ap_skill_new_with_name ("Acrobatics");

    g_test_expect_message ("ap-character",
                           G_LOG_LEVEL_WARNING,
                           "Skill \"*\" does not have a modifier, while not required it is recommended to set a modifier before adding it to the character.");

    retval = ap_character_add_skill (fixture->character, skill);

    g_test_assert_expected_messages ();

    g_assert (retval == TRUE);
}

void
test_character_skill_add_with_no_name (CharacterFixture *fixture,
                                       gconstpointer     user_data)
{
    APSkill *skill;

    skill = ap_skill_new ();

    g_test_expect_message ("ap-character",
                           G_LOG_LEVEL_WARNING,
                           "Skill not registered, skill must have a name to be registered.");

    ap_character_add_skill (fixture->character, skill);

    g_test_assert_expected_messages ();
}

void
test_character_skill_add_with_duplicate_name (CharacterFixture *fixture,
                                              gconstpointer     user_data)
{
    APSkill *skill;

    skill = ap_skill_new_with_name ("Acrobatics");
    ap_skill_set_modifier (skill, "Dexterity");
    ap_character_add_skill (fixture->character, skill);

    skill = ap_skill_new_with_name ("Acrobatics");
    ap_skill_set_modifier (skill, "Dexterity");

    g_test_expect_message ("ap-character",
                           G_LOG_LEVEL_WARNING,
                           "Skill not registered, skill with name \"*\" already exists.");

    ap_character_add_skill (fixture->character, skill);

    g_test_assert_expected_messages ();
}

void
test_character_skill_get (CharacterFixture *fixture,
                          gconstpointer     user_data)
{
    APSkill *skill;
    APSkill *skill_from_character;

    skill = ap_skill_new_with_name ("Acrobatics");

    ap_skill_set_modifier (skill, "Dexterity");

    ap_character_add_skill (fixture->character, skill);

    skill_from_character = ap_character_get_skill (fixture->character, "Acrobatics");

    g_assert (AP_IS_SKILL (skill_from_character));
    g_assert (skill == skill_from_character);
}

void
test_character_remove_ability (CharacterFixture *fixture,
                               gconstpointer     user_data)
{
    APAbility *ability;

    ability = g_object_new (AP_TYPE_ABILITY,
                            "name", "Dexterity",
                            "base", 10,
                            NULL);
    ap_character_add_ability (fixture->character, ability);

    g_assert (ap_character_get_ability (fixture->character, "Dexterity") == ability);

    ap_character_remove_ability (fixture->character, "Dexterity");

    g_assert (ap_character_get_ability (fixture->character, "Dexterity") == NULL);
}

void
test_character_remove_abilities (CharacterFixture *fixture,
                                 gconstpointer     user_data)
{
    APAbility *ability1;
    APAbility *ability2;
    APAbility *ability3;

    ability1 = g_object_new (AP_TYPE_ABILITY,
                             "name", "Dexterity",
                             "base", 10,
                             NULL);
    ap_character_add_ability (fixture->character, ability1);

    ability2 = g_object_new (AP_TYPE_ABILITY,
                             "name", "Strength",
                             "base", 10,
                             NULL);
    ap_character_add_ability (fixture->character, ability2);

    ability3 = g_object_new (AP_TYPE_ABILITY,
                             "name", "Wisdom",
                             "base", 10,
                             NULL);
    ap_character_add_ability (fixture->character, ability3);

    g_assert (ap_character_get_ability (fixture->character, "Dexterity") == ability1);
    g_assert (ap_character_get_ability (fixture->character, "Strength") == ability2);
    g_assert (ap_character_get_ability (fixture->character, "Wisdom") == ability3);

    ap_character_remove_abilities (fixture->character);

    g_assert (ap_character_get_ability (fixture->character, "Dexterity") == NULL);
    g_assert (ap_character_get_ability (fixture->character, "Strength") == NULL);
    g_assert (ap_character_get_ability (fixture->character, "Wisdom") == NULL);
}

void
test_character_remove_skill (CharacterFixture *fixture,
                             gconstpointer     user_data)
{
    APSkill *skill;

    skill = g_object_new (AP_TYPE_SKILL,
                          "name", "Acrobatics",
                          "modifier", "Dexterity",
                          NULL);
    ap_character_add_skill (fixture->character, skill);

    g_assert (ap_character_get_skill (fixture->character, "Acrobatics") == skill);

    ap_character_remove_skill (fixture->character, "Acrobatics");

    g_assert (ap_character_get_skill (fixture->character, "Acrobatics") == NULL);

}

void
test_character_remove_skills (CharacterFixture *fixture,
                              gconstpointer     user_data)
{
    APSkill *skill1;
    APSkill *skill2;
    APSkill *skill3;

    skill1 = g_object_new (AP_TYPE_SKILL,
                           "name", "Acrobatics",
                           "modifier", "Dexterity",
                           NULL);
    ap_character_add_skill (fixture->character, skill1);

    skill2 = g_object_new (AP_TYPE_SKILL,
                           "name", "Animal Handling",
                           "modifier", "Wisdom",
                           NULL);
    ap_character_add_skill (fixture->character, skill2);

    skill3 = g_object_new (AP_TYPE_SKILL,
                           "name", "Athletics",
                           "modifier", "Strength",
                           NULL);
    ap_character_add_skill (fixture->character, skill3);

    g_assert (ap_character_get_skill (fixture->character, "Acrobatics") == skill1);
    g_assert (ap_character_get_skill (fixture->character, "Animal Handling") == skill2);
    g_assert (ap_character_get_skill (fixture->character, "Athletics") == skill3);

    ap_character_remove_skills (fixture->character);

    g_assert (ap_character_get_skill (fixture->character, "Acrobatics") == NULL);
    g_assert (ap_character_get_skill (fixture->character, "Animal Handling") == NULL);
    g_assert (ap_character_get_skill (fixture->character, "Athletics") == NULL);
}

void
test_character_remove_all (CharacterFixture *fixture,
                           gconstpointer     user_data)
{

    APAbility *ability1;
    APAbility *ability2;
    APAbility *ability3;
    APSkill *skill1;
    APSkill *skill2;
    APSkill *skill3;

    g_print ("test1");

    ability1 = g_object_new (AP_TYPE_ABILITY,
                             "name", "Dexterity",
                             "base", 10,
                             NULL);
    ap_character_add_ability (fixture->character, ability1);

    ability2 = g_object_new (AP_TYPE_ABILITY,
                             "name", "Strength",
                             "base", 10,
                             NULL);
    ap_character_add_ability (fixture->character, ability2);

    ability3 = g_object_new (AP_TYPE_ABILITY,
                             "name", "Wisdom",
                             "base", 10,
                             NULL);
    ap_character_add_ability (fixture->character, ability3);

    skill1 = g_object_new (AP_TYPE_SKILL,
                           "name", "Acrobatics",
                           "modifier", "Dexterity",
                           NULL);
    ap_character_add_skill (fixture->character, skill1);

    skill2 = g_object_new (AP_TYPE_SKILL,
                           "name", "Animal Handling",
                           "modifier", "Wisdom",
                           NULL);
    ap_character_add_skill (fixture->character, skill2);

    skill3 = g_object_new (AP_TYPE_SKILL,
                           "name", "Athletics",
                           "modifier", "Strength",
                           NULL);
    ap_character_add_skill (fixture->character, skill3);

    g_assert (ap_character_get_ability (fixture->character, "Dexterity") == ability1);
    g_assert (ap_character_get_ability (fixture->character, "Strength") == ability2);
    g_assert (ap_character_get_ability (fixture->character, "Wisdom") == ability3);

    g_assert (ap_character_get_skill (fixture->character, "Acrobatics") == skill1);
    g_assert (ap_character_get_skill (fixture->character, "Animal Handling") == skill2);
    g_assert (ap_character_get_skill (fixture->character, "Athletics") == skill3);

    g_print ("test");

    ap_character_remove_all (fixture->character);

    g_print ("test");

    g_assert (ap_character_get_ability (fixture->character, "Dexterity") == NULL);
    g_assert (ap_character_get_ability (fixture->character, "Strength") == NULL);
    g_assert (ap_character_get_ability (fixture->character, "Wisdom") == NULL);

    g_assert (ap_character_get_skill (fixture->character, "Acrobatics") == NULL);
    g_assert (ap_character_get_skill (fixture->character, "Animal Handling") == NULL);
    g_assert (ap_character_get_skill (fixture->character, "Athletics") == NULL);
}

void
test_character_signal_ability_base_score_changed_skill_score_changed (CharacterFixture *fixture,
                                                                      gconstpointer     user_data)
{
    APAbility *ability;
    APSkill *skill;

    ability = g_object_new (AP_TYPE_ABILITY,
                            "name", "Dexterity",
                            "base", 10,
                            NULL);
    ap_character_add_ability (fixture->character, ability);

    skill = g_object_new (AP_TYPE_SKILL,
                          "name", "Acrobatics",
                          "modifier", "Dexterity",
                          NULL);
    ap_character_add_skill (fixture->character, skill);

    g_assert (ap_skill_get_score (skill) == 0);

    ap_ability_set_base (ability, 12);

    g_assert (ap_skill_get_score (skill) == 1);
}

void
test_character_signal_skill_modifier_changed (CharacterFixture *fixture,
                                              gconstpointer     user_data)
{
    APAbility *ability;
    APSkill *skill;
    gint modifier1, modifier2, modifier3;

    // NOTE:
    // We don't care if the modifier is generated correctly, ex. ability scores
    // of 10~11 = modifier of 0, ability scores of 12~13 = modifier of 1, etc..
    //
    // We have test for that, for that reason we use ap_ability_get_modifier()
    // just to make sure #APCharacter is sending the right signals, a.k.a the
    // right values.

    // Add Dexterity ability
    ability = ap_ability_new_with_name ("Dexterity");
    ap_ability_set_base (ability, 10);
    modifier1 = ap_ability_get_modifier (ability);
    ap_character_add_ability (fixture->character, ability);

    // Add Strength ability
    ability = ap_ability_new_with_name ("Strength");
    ap_ability_set_base (ability, 12);
    modifier2 = ap_ability_get_modifier (ability);
    ap_character_add_ability (fixture->character, ability);

    // Add Acrobatics skill that uses "Strength" ability
    skill = ap_skill_new_with_name ("Acrobatics");
    ap_skill_set_modifier (skill, "Dexterity");
    ap_character_add_skill (fixture->character, skill);

    // Current "Acrobatics" skill should have a modifier of 0 because an ability
    // score of 10 returns a modifier of 0
    g_assert (ap_skill_get_score (skill) == modifier1);

    // Switch skill to be based of "Strength" ability
    ap_skill_set_modifier (skill, "Strength");

    // Current "Acrobatics" skill should have a modifier of 1 because an ability
    // score of 12 returns a modifier of 1
    g_assert (ap_skill_get_score (skill) == modifier2);

    // Get "Strength" ability as we are currently using it and up it's score to
    // 14. This simulates a user increasing their ability score and changes
    // need to propagate to skills that use "Strength" as a modifier.
    ability = ap_character_get_ability (fixture->character, "Strength");
    ap_ability_set_base (ability, 14);
    modifier3 = ap_ability_get_modifier (ability);

    // Current "Acrobatics" skill should have a modifier of 2 because an ability
    // score of 14 returns a modifier of 2
    g_assert (ap_skill_get_score (skill) == modifier3);
}

void
test_character_check_skill_for_lingering_bind_with_ability (CharacterFixture *fixture,
                                                            gconstpointer     user_data)
{
    APAbility *ability1;
    APAbility *ability2;
    APAbility *character_ability;
    APSkill *skill;

    ability1 = g_object_new (AP_TYPE_ABILITY,
                             "name", "Dexterity",
                             "base", 10,
                             NULL);
    ap_character_add_ability (fixture->character, ability1);

    ability2 = g_object_new (AP_TYPE_ABILITY,
                             "name", "Strength",
                             "base", 12,
                             NULL);
    ap_character_add_ability (fixture->character, ability2);

    skill = g_object_new (AP_TYPE_SKILL,
                          "name", "Acrobatics",
                          "modifier", "Dexterity",
                          NULL);
    ap_character_add_skill (fixture->character, skill);

    // Connecting to an existing ability should get that's ability modifier
    // and update the Skills score
    g_print ("skill0: %i\n", ap_skill_get_score (skill));
    g_assert (ap_skill_get_score (skill) == 0);

    ap_skill_set_modifier (skill, "Strength");

    // Switching skill should issue new notify:modifier & get updated
    // #APAbility:score
    g_print ("skill1: %i\n", ap_skill_get_score (skill));
    g_assert (ap_skill_get_score (skill) == 1);

    character_ability = ap_character_get_ability (fixture->character, ap_skill_get_modifier (skill));

    // Skill should be returning new updated "modifier"
    g_assert (character_ability == ability2);

    ap_ability_set_base (ability1, 14);

    // Updating the old ability connected to the Skill should **NOT** update
    // the score!
    g_print ("skill2: %i\n", ap_skill_get_score (skill));
    g_assert (ap_skill_get_score (skill) == 1);
}
